# -*- coding: utf-8 -*-

"""

Module for downloading Sentinel-2 images using EODAG
https://www.github.com/CS-SI/EODAG

"""

import logging
import shapefile
import os
import shutil
from pathlib import Path
from eodag import EODataAccessGateway
from eodag import setup_logging
from queue import Queue
from eodag.utils.logging import get_logging_verbose
from threading import Thread
import multiprocessing

from .config import SHARED_DATA, Config
from .utils import get_tile

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

ROOT = Path(os.path.realpath(__file__)).parent.parent

class S2cEodag:

    def __init__(
        self, 
        name: str = None,
        provider: str = "peps"
    ):
    
        self.name = name
        self.provider = provider
        self.products = None
        eodag_centroids_shp = SHARED_DATA.get("eodag_centroids_shp")
        
        with shapefile.Reader(eodag_centroids_shp) as shp:
            shaperecs = shp.shapeRecords()
        
        locations_yaml_content = """
        shapefiles:
          - name: s2_tile_centroid
            path: {}
            attr: tile_id
        """.format(os.path.abspath(eodag_centroids_shp))
        
        eodag_custom_location_cfg = os.path.abspath(Config()._CONFIG_DIR / "eodag_custom_locations.yml")

        with open(eodag_custom_location_cfg, "w") as f_yml:
            f_yml.write(locations_yaml_content.strip())
        logging.disable(level=logging.WARNING)
        self.dag = EODataAccessGateway(locations_conf_path = eodag_custom_location_cfg)
        logging.disable(logging.NOTSET)

        self.dag.set_preferred_provider(self.provider)
        # logger.info(self.dag.available_providers("S2_MSI_L1C"))

        targeted_tile = [
            sr
            for sr in shaperecs
            if sr.record["tile_id"] == name
        ][0]
        
    
    def search(
        self,
        productType: str = "L1C",
        start: str = "2015-01-01",
        end: str = "9999-12-31",
        ref: str = "l1c",
        min_cloudcover: int = 0,
        max_cloudcover: int = 100,    
    ):
        ######### ici faudrait virer le self.productType, qui ne doit pas être global pour le download... 
        if productType == "L1C":
            self.productType = "S2_MSI_L1C"
            outputs_prefix = Path(Config().get("l1c_path")) / self.name
        elif productType == "L2A":
            self.productType = "S2_MSI_L2A"
            outputs_prefix = Path(Config().get("l2a_path")) / self.name

        default_search_criteria = dict(
            productType = self.productType,
            start = start,
            end = end, 
            # items_per_page=500,
        )
        
        logging.disable(level=logging.WARNING)
        self.products = self.dag.search_all(
            locations = dict(s2_tile_centroid = self.name),
            **default_search_criteria
        )
        logging.disable(logging.NOTSET)
        
        fitered = self.products[:]
        for p in fitered:
            if (outputs_prefix / (p.properties["title"] + ".SAFE")).exists():
                l1c_presence = "PRESENT"
            else:
                l1c_presence = "ABSENT"
            if (Path(Config().get("l2a_path")) / self.name / (p.properties["title"].replace("L1C_", "L2A_").replace("__OPER__", "_USER_") + ".SAFE")).exists():
                l2a_presence = "PRESENT"
            else:
                l2a_presence = "ABSENT"
            
            ##### remonter la condition l1c
            
            if self.name not in p.properties["title"]:
                self.products.remove(p)
                logger.info("{} - local l1c {} - filtering (wrong Tile)".format(p.properties["title"], l1c_presence))
            else:
                # logger.info(p.properties["cloudCover"])
                if not(min_cloudcover <= float(p.properties["cloudCover"] or 0) <= max_cloudcover):
                    self.products.remove(p)
                    logger.info("{} - local l1c {} - filtering (CC = {}%)".format(p.properties["title"], l1c_presence, int(p.properties["cloudCover"])))
                else:
                    if ref == "l1c":
                        if (outputs_prefix / (p.properties["title"] + ".SAFE")).exists():
                            logger.info("{} - local l1c {} - filtering".format(p.properties["title"], l1c_presence, p.properties["storageStatus"]))
                            self.products.remove(p)
                        else:
                            logger.info("{} - local l1c {} - remote {}".format(p.properties["title"], l1c_presence, p.properties["storageStatus"]))
                    elif ref == "l2a":
                        if (Path(Config().get("l2a_path")) / self.name / (p.properties["title"].replace("L1C_", "L2A_").replace("__OPER__", "_USER_") + ".SAFE")).exists():
                            logger.info("{} - local l1c {} - local l2a {} - filtering".format(p.properties["title"], l1c_presence, l2a_presence, p.properties["storageStatus"]))
                            self.products.remove(p)
                        else:
                            if (outputs_prefix / (p.properties["title"] + ".SAFE")).exists():
                                logger.info("{} - local l1c {} - filtering".format(p.properties["title"], l1c_presence, p.properties["storageStatus"]))
                                self.products.remove(p)
                            else:
                                logger.info("{} - local l1c {} - local l2a {} - remote {}".format(p.properties["title"], l1c_presence, l2a_presence, p.properties["storageStatus"]))
                            
                            
                    
                    
                      
        # clouds = self.products[:]
        # for p in clouds:
            # if not(min_cloudcover <= int(p.properties["cloudCover"]) <= max_cloudcover):
                # self.products.remove(p)
                # logger.info("{} - wrong cloud cover ({}%) - filtering".format(p.properties["title"], int(p.properties["cloudCover"])))        
        # local = self.products[:]
        # for p in local:
            # if ref == "l1c":
                # if (outputs_prefix / (p.properties["title"] + ".SAFE")).exists():
                    # logger.info("{} - remote {} - local l1c PRESENT - filtering".format(p.properties["title"], p.properties["storageStatus"]))
                    # self.products.remove(p)
                # else:
                    # logger.info("{} - remote {} - local l1c ABSENT".format(p.properties["title"], p.properties["storageStatus"]))
            # elif ref == "l2a":
                # if (Path(Config().get("l2a_path")) / self.name / (p.properties["title"].replace("L1C_", "L2A_").replace("__OPER__", "_USER_") + ".SAFE")).exists():
                    # logger.info("{} - remote {} - local l2a PRESENT - filtering".format(p.properties["title"], p.properties["storageStatus"]))
                    # self.products.remove(p)
                # else:
                    # logger.info("{} - remote {} - local l2a ABSENT".format(p.properties["title"], p.properties["storageStatus"]))
        logger.info("Search returned {} new product(s) to download".format(len(self.products))) ####### rajouter ici "- dont xx ABSENT - dont xx ONLINE / xx STAGING"
            
    def download(
        self,
        product_id, 
        outputs_prefix: str = None,
        extract: bool = True,
        delete_archive: bool = True,
    ):
        if not outputs_prefix:
            if "L1C" in product_id.properties['title']:
                root_path = "l1c_path"
            elif "L2A" in product_id.properties['title']:
                root_path = "l2a_path"
            outputs_prefix = str(Path(Config().get(root_path)) / self.name)
        
        setup_logging(verbose = 2)

        if product_id.properties["storageStatus"] == "ONLINE":
            try:
                # logging.disable(level=logging.WARNING)
                downloaded_path = self.dag.download(
                    product_id, 
                    outputs_prefix = outputs_prefix, 
                    extract = extract, 
                    delete_archive = delete_archive,
                    wait=1,
                    timeout=0,
                )
                # logging.disable(logging.NOTSET)
                if Path(downloaded_path).stem == Path(downloaded_path).parent.stem:
                    upper_location = Path(downloaded_path).parent.parent / Path(downloaded_path).name
                    Path(downloaded_path).replace(upper_location)
                    product_id.location = "file://" + str(upper_location)
                    Path(downloaded_path).parent.rmdir()
                    logger.info("Moving up SAVE file")
                # modification de l'arborescence si il y a un sous dossier en trop (provider onda par ex)
                if os.listdir(downloaded_path)[0] == str(Path(downloaded_path).name):
                    shutil.copytree(downloaded_path + '/' + os.listdir(downloaded_path)[0],downloaded_path,dirs_exist_ok=True)
                    logger.info("Moving up files")
                    shutil.rmtree(downloaded_path + '/' + os.listdir(downloaded_path)[0])
                    logger.info("Deleting old folder")
            except:
                logger.info("Error: ONLINE product but cannot be downloaded, check your credentials or retry later")

        elif product_id.properties["storageStatus"] == "OFFLINE":
            try:
                setup_logging(verbose = 0)
                downloaded_path = self.dag.download(
                    product_id, 
                    outputs_prefix = outputs_prefix, 
                    extract = extract, 
                    delete_archive = delete_archive,
                    wait=1,
                    timeout=0,
                )
            except:
                logger.info("remote OFFLINE, ordered")

        elif product_id.properties["storageStatus"] == "STAGING":
            logger.info("remote STAGING, retry later")

    def order_offline(
        self,
        product_id, 
    ):
        setup_logging(verbose = 0)
        if product_id.properties["storageStatus"] == "OFFLINE":
            try:
                downloaded_path = self.dag.download(
                    product_id, 
                    wait=1,
                    timeout=0,
                )
            except:
                logger.info("{} - remote OFFLINE, ordered".format(product_id.properties["title"]))
        setup_logging(verbose = 2)

def S2cEodag_download_uniq(
    product, 
    outputs_prefix: str = None,
    extract: bool = True,
    delete_archive: bool = True,
):
    tile_name = get_tile(product.properties['title'])
    if not outputs_prefix:
        if "L1C" in product.properties['title']:
            root_path = "l1c_path"
        elif "L2A" in product.properties['title']:
            root_path = "l2a_path"
        outputs_prefix = str(Path(Config().get(root_path)) / tile_name)
    
    setup_logging(verbose = 2)
    
    logging.disable(level=logging.WARNING)
    dag = EODataAccessGateway()
    logging.disable(logging.NOTSET)

    if product.properties["storageStatus"] == "ONLINE":
        try:
            # logging.disable(level=logging.WARNING)
            downloaded_path = dag.download(
                product, 
                outputs_prefix = outputs_prefix, 
                extract = extract, 
                delete_archive = delete_archive,
                wait=1,
                timeout=0,
            )
            # logging.disable(logging.NOTSET)

            if Path(downloaded_path).stem == Path(downloaded_path).parent.stem:
                upper_location = Path(downloaded_path).parent.parent / Path(downloaded_path).name
                Path(downloaded_path).replace(upper_location)
                product.location = "file://" + str(upper_location)
                Path(downloaded_path).parent.rmdir()
                logger.info("Moving up SAVE file")
        except:
            logger.info("Error: ONLINE product but cannot be downloaded, retry later")

    elif product.properties["storageStatus"] == "OFFLINE":
        try:
            setup_logging(verbose = 0)
            downloaded_path = dag.download(
                product, 
                outputs_prefix = outputs_prefix, 
                extract = extract, 
                delete_archive = delete_archive,
                wait=1,
                timeout=0,
            )
        except:
            logger.info("OFFLINE product, ordered, retry later")

    elif product.properties["storageStatus"] == "STAGING":
        logger.info("STAGING product, retry later")            
    
    
def S2cEodag_download(
    download_list: list = None,
    dl_mode: str = "multit",
    outputs_prefix: str = None,
    extract: bool = True,
    delete_archive: bool = True,
):
    
    ## sequential
    if dl_mode == "seq":  
        for product in download_list:
            S2cEodag_download_uniq(
                product,
                outputs_prefix = outputs_prefix, 
                extract = extract,
                delete_archive = delete_archive,                                
            )     
            
    ## multithreading
    elif dl_mode == "multit":
        NUM_THREADS = 8
        q = Queue()
        def do_stuff(q):
          while True:
            S2cEodag_download_uniq(q.get())
            q.task_done()
        for product in download_list:
            q.put(product)
        for t in range(NUM_THREADS):
            worker = Thread(target = do_stuff, args = (q,))
            worker.daemon = True
            worker.start()
        q.join()
        
    ## multiprocessing
    elif dl_mode == "multip":
        nb_proc = 8
        nb_proc = max(min(len(os.sched_getaffinity(0)) - 1, nb_proc), 1)
        pool = multiprocessing.Pool(nb_proc)
        results = [pool.map(S2cEodag_download_uniq, download_list)]
        pool.close()
        pool.join()



