# coding: utf-8

"""
Module for managing sen2chain processing jobs
"""

import logging
from pathlib import Path
import pandas as pd
import datetime
from itertools import chain
import re
import setuptools
from crontab import CronTab
from collections import OrderedDict
from configparser import ConfigParser


from .config import SHARED_DATA, Config
from .indices import IndicesCollection
from .library import Library
from .tiles import Tile
from .utils import datetime_to_str
from .multi_processing import (
    l2a_multiprocessing,
    idx_multiprocessing,
    cld_version_probability_iterations_reprocessing_multiprocessing,
)
from .download_eodag import (
    S2cEodag_download, 
    S2cEodag_download_uniq,
)

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Jobs:
    """
    Class to manage created jobs
    """

    def __init__(self):
        self._jobs_all = list(Config()._JOBS_DIR.glob("job_*.cfg"))
        self._jobs_all.extend(list(Config()._JOBS_DIR.glob("job_*.py")))
        self._jid_set = set([t.stem[4:] for t in self._jobs_all])
        self._config_files_exist = []
        self._py_scripts_exist = []
        self._logs = []
        self._timing = []
        self._cron_status = []
        self._cron_timing = []
        for jid in self._jid_set:
            logger.disabled = True
            j = Job(jid)
            self._config_files_exist.append(j._config_path.exists())
            self._py_scripts_exist.append(j._python_script_path.exists())
            self._logs.append(j.logs)
            self._timing.append(j.timing)
            self._cron_status.append(j.cron_status)
            self._cron_timing.append(j.cron_timing)
            logger.disabled = False

    def __repr__(self):
        if self._jobs_all:
            return repr(
                pd.DataFrame(
                    OrderedDict(
                        [
                            ("job_id", list(self._jid_set)),
                            ("config_file", self._config_files_exist),
                            ("python_script", self._py_scripts_exist),
                            ("logging", self._logs),
                            ("timing", self._timing),
                            ("cron_status", self._cron_status),
                            ("cron_timing", self._cron_timing),
                        ]
                    )
                )
            )
        else:
            logger.info("No Jobs here")
            return repr(None)

        # >>> job.render().split(' ')[0:4]

    @property
    def list(self):
        return [t.stem[4:] for t in self._jobs_all]

    def remove(self, jid: str = None):
        if jid in self.list:
            logger.disabled = True
            j = Job(jid)
            logger.disabled = False
            j.cron_remove()
            if j._python_script_path.exists():
                j._python_script_path.unlink()
                logger.info("Removing Python script...")
            if j._config_path.exists():
                j._config_path.unlink()
                logger.info("Removing config file...")
        else:
            logger.info("Job {} not found, doing nothing...".format(jid))


class Job:
    """
    Class to manage job
    """

    # logger.propagate = False
    def __init__(self, jid: str):
        self._config_path = Config()._JOBS_DIR / ("job_" + jid + ".cfg")
        self._python_script_path = Config()._JOBS_DIR / ("job_" + jid + ".py")
        self.jid = jid
        if self._config_path.exists():
            logger.info("Reading existing config...")
            self.read(self._config_path)
        else:
            logger.info("Creating new job...")
            self.init()
        self._cron = CronTab(user=True)
        self.cron_status = self.get_cron_status()

        self._log_folder_path = Path(Config().get("log_path")) / ("job_" + jid)

    def __repr__(self):
        return repr(self.tasks)

    def init(self):
        raw_job_cfg = SHARED_DATA.get("raw_job_cfg")
        self.read(raw_job_cfg)
        first_row = OrderedDict(
            [
                ("tile", ["40KCB"]),
                ("date_min", [datetime_to_str(datetime.datetime.now(), date_format = 'ymd')]),
                ("date_max", [datetime_to_str(datetime.datetime.now(), date_format = 'ymd')]),
                ("max_clouds", [100]),
                ("l1c", [False]),
                ("l2a", [False]),
                ("cloudmask", [False]),
                ("indices", [False]),
                ("remove", [False]),
                ("comments", [""]),
            ]
        )
        self.tasks = pd.DataFrame(first_row)

    def task_add(self, row: dict = None):
        if not row:
            logger.info("No row provided, using default")
            row = pd.DataFrame(
                {
                    "tile": ["40KCB"],
                    "date_min": [datetime_to_str(datetime.datetime.now(), date_format = 'ymd')],
                    "date_max": [datetime_to_str(datetime.datetime.now(), date_format = 'ymd')],
                    "max_clouds": [100],
                    "l1c": [False],
                    "l2a": [False],
                    "cloudmask": [False],
                    "indices": [False],
                    "remove": [False],
                    "comments": [""],
                }
            )
        # self.tasks = self.tasks.append(row, ignore_index=True)
        self.tasks = pd.concat([self.tasks, row], ignore_index=True)[
            self.tasks.columns
        ]

        logger.info("\n{}".format(self.tasks))

    def task_edit(self, task_id: int = None, **kwargs):
        if task_id is None:
            logger.info(
                "Please provide task_number to edit, if no task in job, create_task first"
            )
        else:
            if task_id in self.tasks.index:
                for arg, val in kwargs.items():
                    if arg in self.tasks.columns:
                        self.tasks.loc[task_id, arg] = val
                        logger.info(
                            "Line {}: {} updated to {}".format(
                                task_id, arg, val
                            )
                        )
                    else:
                        logger.info("{} not found".format(arg))
                self.clean_fields()
                logger.info("\n{}".format(self.tasks))
            else:
                logger.info("Task_number not found")

    def task_remove(self, task_id: int = None):
        if task_id is None:
            logger.info("Please provide task_number to remove")
        else:
            if task_id in self.tasks.index:
                self.tasks.drop(labels=task_id, axis=0, inplace=True)
                self.tasks.reset_index(inplace=True, drop=True)
                logger.info("\n{}".format(self.tasks))
            else:
                logger.info("Task_number not found")

    def save(self):
        # save task to disk
        with open(str(self._config_path), "w") as ict:
            comments_header ="\n".join(
                [
                    "### Parameters", 
                    "# logs: True | False", 
                    "# timing: in cron format", 
                    "# tile: tile identifier, format ##XXX, comment line using ! before tile name"
                    "# l1c: download l1c: True|False",
                    "# l2a: compute l2a with sen2chain: True | False", 
                    "# cloudmasks: False | CM001 | CM002 | CM003-PRB1-ITER5 | CM004-CSH1-CMP1-CHP1-TCI1-ITER0",
                    "# indices: False | All | NDVI/NDWIGAO/etc.", 
                    "# remove: False | l1c/l2a",
                    "# comments: free user comments, ie tile name, etc.", 
                    "",
                    "",
                ]
            )
            header = "\n".join(
                [
                    "logs = " + str(self.logs),
                    "timing = " + self.timing,
                    "",
                    "",
                ]
            )
            
            for line in comments_header:
                ict.write(line)
            
            for line in header:
                ict.write(line)
            # self.tasks.to_csv(ict)
            self.unclean_fields()
            self.tasks.to_csv(ict, index=False, sep=";")

    def get_cron_status(self):
        iter = list(self._cron.find_comment("sen2chain_job_" + self.jid))
        if iter:
            for job in iter:
                if job.is_enabled():
                    status = "enabled"
                else:
                    status = "disabled"
                self.cron_job = job
                self.cron_timing = str(self.cron_job.slices)
        else:
            status = "absent"
            self.cron_job = None
            self.cron_timing = None
        return status

    def create_python_script(self):
        lines = ["# -*- coding:utf-8 -*-\n"]
        lines.append("\n")
        lines.append('"""\n')
        lines.append("Module for managing sen2chain processing jobs\n")
        lines.append('"""\n')
        lines.append("\n")
        lines.append("from sen2chain import Job\n")
        lines.append(
            'Job("'
            + self.jid
            + '").run(clean_before = False, clean_after = False)\n'
        )
        with open(str(self._python_script_path), "w") as f:
            f.writelines(lines)

    def cron_enable(
        self,
    ):
        # enable job in cron
        self.save()
        self.create_python_script()
        iter = list(self._cron.find_comment("sen2chain_job_" + self.jid))
        if iter:
            for job in iter:
                logger.info("Enabling job")
                if self.timing:
                    job.setall(self.timing)
                job.enable()
        else:
            job = self._cron.new(
                command="/usr/bin/python3 " + str(self._python_script_path),
                comment="sen2chain_job_" + self.jid,
            )
            if self.timing:
                job.setall(self.timing)
            else:
                job.setall("0 0 * * *")
            job.enable()
            logger.info("Enabling job...")

        # logger.info("Time: {}".format(job.time))
        self._cron.write()
        # new.enable(False)
        self.get_cron_status()

    def cron_disable(self):
        # disable / commenting job in cron
        iter = list(self._cron.find_comment("sen2chain_job_" + self.jid))
        if iter:
            for job in iter:
                logger.info("Disabling job...")
                job.enable(False)
            self._cron.write()
            self.get_cron_status()

    def cron_remove(self):
        # remove job from cron
        iter = list(self._cron.find_comment("sen2chain_job_" + self.jid))
        if iter:
            for job in iter:
                logger.info("Removing job from cron...")
                self._cron.remove(job)
            self._cron.write()
            self.get_cron_status()

    def read(self, path):
        parser = ConfigParser(allow_no_value=True, strict = False)
        with open(str(path)) as stream:
            parser.read_string(
                "[top]\n" + stream.read(),
            )  # This line does the trick.
        self.logs = bool(setuptools.distutils.util.strtobool(parser["top"]["logs"]))
        self.timing = parser["top"]["timing"]

        self.tasks = pd.read_csv(
            path,
            sep=";",
            na_values="",
            na_filter=False,
            comment="#",
            dtype=str,
            header=2,
        )
        
        # if "logs" not in self.tasks:
        # self.tasks["logs"] = False
        
        ##### prévoir ici un check pour les champs nouveaux ou manquants !
        if 'comments' not in self.tasks:
            self.tasks['comments'] = ""
        self.clean_fields()
    
    def clean_fields(self):
        for index, row in self.tasks.iterrows():
            if not row.date_min:
                # self.tasks.at[index, "start_time"] = (datetime.datetime.now()-datetime.timedelta(days=delta_t)).strftime('%Y-%m-%d')
                self.tasks.at[index, "date_min"] = datetime.datetime.strptime(
                    "2015-01-01", "%Y-%m-%d"
                ).strftime("%Y-%m-%d")
            if not row.date_max:
                self.tasks.at[index, "date_max"] = (
                    datetime.datetime.now() + datetime.timedelta(days=1)
                ).strftime("%Y-%m-%d")
            if not row.max_clouds:
                self.tasks.at[index, "max_clouds"] = 100
            if not row.indices == "False":
                if row.indices == "All":
                    self.tasks.at[
                        index, "indices"
                    ] = IndicesCollection.available_indices
                else:
                    if not isinstance(self.tasks.at[index, "indices"] , list):
                        self.tasks.at[index, "indices"] = str(row.indices).split(
                            "/"
                        )

    def unclean_fields(self):
        for index, row in self.tasks.iterrows():
            if (not row.indices == "False") and (not row.indices == "All"):
                    self.tasks.at[index, "indices"] = '/'.join(self.tasks.at[index, "indices"])   
    
    @staticmethod
    def get_cm_version(identifier) -> str:
        """Returns cloudmask version from a cloud mask identifier string.
        :param string: string from which to extract the version name.
        """
        try:
            pat = re.compile(r"(?P<cm_version>CM00[1-2])")
            # return pat.match(identifier).groupdict()
        except:
            pass
        try:
            pat = re.compile(
                r"(?P<cm_version>CM003)"
                + "-PRB(?P<probability>.*)"
                + "-ITER(?P<iterations>.*)"
            )
            # return pat.match(identifier).groupdict()
        except:
            pass
        try:
            pat = re.compile(
                r"(?P<cm_version>CM004)"
                + "-CSH(?P<cld_shad>.*)"
                + "-CMP(?P<cld_med_prob>.*)"
                + "-CHP(?P<cld_hi_prob>.*)"
                + "-TCI(?P<thin_cir>.*)"
                + "-ITER(?P<iterations>.*)"
            )
            # return pat.match(identifier).groupdict()
        except:
            pass
        try:
            returned_val = pat.match(identifier).groupdict()
            if "probability" not in returned_val:
                returned_val["probability"] = 1
            if "iterations" not in returned_val:
                returned_val["iterations"] = 5
            if "cld_shad" not in returned_val:
                returned_val["cld_shad"] = True
            if "cld_med_prob" not in returned_val:
                returned_val["cld_med_prob"] = True
            if "cld_hi_prob" not in returned_val:
                returned_val["cld_hi_prob"] = True
            if "thin_cir" not in returned_val:
                returned_val["thin_cir"] = True
        except:
            returned_val = {
                "cm_version": "cm001",
                "probability": 1,
                "iterations": 5,
                "cld_shad": True,
                "cld_med_prob": True,
                "cld_hi_prob": True, 
                "thin_cir": True, 
            }
        
        
        return returned_val


    def run(
        self,
        nb_proc: int = 12,
        clean_before: bool = False,
        clean_after: bool = False,
    ):
        
        if self.logs:
            self._log_folder_path.mkdir(exist_ok=True)
            self._log_file = self._log_folder_path / (
                "job_"
                + self.jid
                + "_run_"
                + datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
                + ".log"
            )
            f = open(str(self._log_file), "w")
            f.write("Start : {}\n\n".format(datetime.datetime.now()))
            f.write(repr(self) + "\n\n")
            f.flush()
        
        tasks = self.tasks
        tasks = tasks[tasks["tile"].str.contains("!") == False]
        
        if not tasks.empty:
            ## Cleaning before
            if clean_before:
                logger.info("Cleaning Tiles")
                if self.logs:
                    f.write("Cleaning Tiles\n")
                    f.flush()
                
                clean_list = []
                for index, row in tasks.iterrows():
                    clean_list.append(row.tile)
                lib = Library()
                pb_before = lib.clean(
                    clean_list, remove=True, remove_indice_tif=True
                )
                # lib.clean(clean_list, remove=False)
            
            ## L1C download - each tile sequential
            # download_list= []
            # logger.info("Downloading l1c seq")
            # for index, row in tasks.iterrows():
                # if bool(setuptools.distutils.util.strtobool(str(row.l1c))):
                    # t = Tile(row.tile)
                    # logger.info("Tile: {}".format(t.name))
                    # downloaded = t.get_l1c(
                        # provider = "peps",
                        # download = True,
                        # dl_mode = "multit",
                        # start = row.date_min,
                        # end = row.date_max,
                        # new = False,
                    # )
                    # logger.info("Downloaded list: {}".format(downloaded))
                    # download_list.extend(downloaded)
            # logger.info("download_list: {}".format(download_list))

            ## L1C download - all tiles multi
            download_list= []
            logger.info("Downloading l1c products")
            if self.logs:
                f.write("Downloading l1c products\n")
                f.flush()

            for index, row in tasks.iterrows():
                if bool(setuptools.distutils.util.strtobool(str(row.l1c))):
                    t = Tile(row.tile)
                    logger.info("Tile: {}".format(t.name))
                    if self.logs:
                        f.write("Tile: {}\n".format(t.name))
                        f.flush()

                    tile_download_list = t.get_l1c(
                        provider = "peps",
                        download = False,
                        start = row.date_min,
                        end = row.date_max,
                        new = False,
                    )
                    download_list.extend(tile_download_list)

            ### download 
            before_list = [p.location for p in download_list]
            # S2cEodag_download(download_list, "seq")
            S2cEodag_download(download_list, "multit")
            # S2cEodag_download(download_list, "multip")
            after_list = [p.location for p in download_list]
            downloaded_products = [Path(after_list[i]).stem for i in range(len(before_list)) if before_list[i] != after_list[i]]
            logger.info("Downloaded product(s): {}".format(downloaded_products))
            if self.logs:
                f.write("{} downloaded product(s): {}\n\n".format(len(downloaded_products), downloaded_products))
                f.flush()
    
            # Traitement des L1C en L2A
            logger.info("Computing l2a")
            if self.logs:
                f.write("Computing l2a\n")
                f.flush()

            l1c_process_list = []
            for index, row in tasks.iterrows():
                if bool(setuptools.distutils.util.strtobool(str(row.l2a))):
                    t = Tile(row.tile)
                    l1c_to_process = list(
                        p.identifier
                        for p in t.l2a_missings.filter_dates(
                            date_min=row.date_min, date_max=row.date_max
                        )
                    )
                    l1c_process_list.append(l1c_to_process)
                    logger.info(
                        "ajout {}: {} l1c files\n".format(
                            row.tile, len(l1c_to_process)
                        )
                    )
            l1c_process_list = list(chain.from_iterable(l1c_process_list))
            logger.info(
                "Process list ({} files): {}".format(
                    len(l1c_process_list), l1c_process_list
                )
            )
            if self.logs:
                f.write(
                    "Process list ({} files): {}\n\n".format(
                        len(l1c_process_list), l1c_process_list
                    )
                )
                f.flush()
            
            l2a_res = False
            if l1c_process_list:
                l2a_res = l2a_multiprocessing(
                    l1c_process_list, nb_proc=nb_proc
                )

            # Remove downloaded L1C
            for index, row in tasks.iterrows():
                if "l1c" in str(row.remove).lower():
                    t = Tile(row.tile)
                    prodlist = [p for p in downloaded_products if row.tile in p]
                    t.remove_l1c(prodlist)
                    logger.info("Removing downloaded l1c products: {}".format(prodlist))
                    if self.logs:
                        f.write("Removing downloaded l1c products: {}\n\n".format(prodlist))
                        f.flush()                    

            # Comuting cloudmasks (from L2A)
            logger.info("Computing cloudmasks")
            if self.logs:
                f.write("Computing cloudmasks\n")
                f.flush()
            reprocess = False
            cld_l2a_process_list = []
            for index, row in tasks.iterrows():
                # if not bool(distutils.util.strtobool(str(row.cloudmask))):
                if not (row.cloudmask == "False" or not row.cloudmask):
                    cloudmask = self.get_cm_version(row.cloudmask)                    
                    t = Tile(row.tile)
                    l2a_to_process = [
                        p.identifier
                        for p in t.cloudmasks_missing().filter_dates(
                            date_min=row.date_min, date_max=row.date_max
                        )
                    ]
                    for j in l2a_to_process:
                        l2a_cm_details = [
                            j,
                            cloudmask["cm_version"],
                            cloudmask["probability"],
                            cloudmask["iterations"],
                            cloudmask["cld_shad"],
                            cloudmask["cld_med_prob"],
                            cloudmask["cld_hi_prob"],
                            cloudmask["thin_cir"],
                            reprocess,
                        ]
                        cld_l2a_process_list.append(l2a_cm_details)
                    logger.info(
                        "{} adding: {} l2a products to process".format(
                            row.tile, len(l2a_to_process)
                        )
                    )
                    if self.logs:
                        f.write(
                            "{} adding: {} l2a products to process\n".format(
                                row.tile, len(l2a_to_process)
                            )
                        )
                        f.flush()
                    
            logger.info(
                "Cloudmask computing process list ({} l2a products): {}".format(
                    len(cld_l2a_process_list), cld_l2a_process_list
                )
            )
            if self.logs:
                f.write(
                    "Cloudmask computing process list ({} l2a products): {}\n\n".format(
                        len(cld_l2a_process_list), cld_l2a_process_list
                    )
                )
                f.flush()
                
            cld_res = False
            if cld_l2a_process_list:
                cld_res = cld_version_probability_iterations_reprocessing_multiprocessing(
                    cld_l2a_process_list, nb_proc=nb_proc
                )

            # Traitement des L2A (indices)
            logger.info("Computing indices")
            if self.logs:
                f.write("Computing indices\n")
                f.flush()
            nodata_clouds = True
            quicklook = False
            indices_l2a_process_list = []
            for index, row in tasks.iterrows():
                if not (row.indices == "False" or not row.indices):
                    t = Tile(row.tile)
                    # indices_list = row.indices.split("/")
                    for i in row.indices:
                        # l2a_list = [p.identifier for p in t.missing_indices(i).filter_dates(date_min = row.start_time, date_max = row.end_time)]
                        l2a_to_process = [
                            p.identifier
                            for p in t.missing_indices(
                                indice=i,
                                nodata_clouds=nodata_clouds,
                            ).filter_dates(
                                date_min=row.date_min, date_max=row.date_max
                            )
                        ]
                        for j in l2a_to_process:
                            l2a_ind_details = [
                                j,
                                i,
                                reprocess,
                                nodata_clouds,
                                quicklook,
                                cloudmask["cm_version"],
                                cloudmask["probability"],
                                cloudmask["iterations"],
                                cloudmask["cld_shad"],
                                cloudmask["cld_med_prob"],
                                cloudmask["cld_hi_prob"],
                                cloudmask["thin_cir"],
                            ]
                            indices_l2a_process_list.append(l2a_ind_details)
                        logger.info(
                            "{}: adding - {}: {} l2a products".format(
                                row.tile, i, len(l2a_to_process)
                            )
                        )
                        if self.logs:
                            f.write(
                            "{}: adding - {}: {} l2a products".format(
                                row.tile, i, len(l2a_to_process)
                                )
                            )
                            f.flush()
            logger.info(
                "Indice computing process list ({} l2a products): {}".format(
                    len(indices_l2a_process_list), indices_l2a_process_list
                )
            )
            if self.logs:
                f.write(
                    "Indice computing process list ({} l2a products): {}\n\n".format(
                        len(indices_l2a_process_list), indices_l2a_process_list
                    )
                )
                f.flush()            
            
            indices_res = False
            if indices_l2a_process_list:
                indices_res = idx_multiprocessing(
                    indices_l2a_process_list, nb_proc=nb_proc
                )

            # Remove L2A
            # todo

            # Cleaning after
            if clean_after:
                logger.info("Cleaning Tiles")
                if self.logs:
                    f.write("Cleaning Tiles\n\n")
                    f.flush()
                
                clean_list = []
                for index, row in tasks.iterrows():
                    clean_list.append(row.tile)
                lib = Library()
                pb_before = lib.clean(
                    clean_list, remove=True, remove_indice_tif=True
                )
                # lib.clean(clean_list, remove=False)

        else:
            logger.info("No task defined for this job, doing nothing")
            if self.logs:
                f.write("No task defined for this job, doing nothing\n\n")
                f.flush()

        if self.logs:
            f.write("Fin : {}\n".format(datetime.datetime.now()))
            f.close()
